libpath = '/home/l_samenv/frappy/cetoniSDK/CETONI_SDK_Raspi_64bit_v20220627/python/src/'
import sys
if libpath not in sys.path:
    sys.path.append(libpath)

from frappy.core import Drivable, Readable, StringIO, HasIO, FloatRange, IntRange, StringType, BoolType, EnumType, \
    Parameter, Property, PersistentParam, Command, IDLE, BUSY, ERROR, WARN, Attached, Module
from qmixsdk import qmixbus
from qmixsdk import qmixpump
from qmixsdk import qmixvalve
from qmixsdk.qmixpump import ContiFlowProperty, ContiFlowSwitchingMode
from qmixsdk.qmixbus import UnitPrefix, TimeUnit
import time

class LabCannBus(Module):
    deviceconfig = Property('config files', StringType(),default="/home/l_samenv/frappy/cetoniSDK/CETONI_SDK_Raspi_64bit_v20220627/config/dual_pumps")

    def earlyInit(self):
        super().earlyInit()
        self.bus = qmixbus.Bus()
        self.bus.open(self.deviceconfig, "")

    def initModule(self):
        super().initModule()
        self.bus.start()

        with open('/sys/class/ionopimax/buzzer/beep', 'w') as f :
            f.write('200 50 3')

    def shutdownModule(self):
        """Close the connection"""
        self.bus.stop()
        self.bus.close()


class SyringePump(Drivable):
    io = Attached()
    pump_name = Property('name of pump', StringType(),default="Nemesys_S_1_Pump")
    valve_name = Property('name of valve', StringType(),default="Nemesys_S_1_Valve")
    
    inner_diameter_set = Property('inner diameter', FloatRange(), default=1)
    piston_stroke_set = Property('piston stroke', FloatRange(), default=60)
    
    value = Parameter('volume', FloatRange(unit='uL'))
    status = Parameter()

    max_flow_rate = Parameter('max flow rate', FloatRange(0,100000, unit='uL/s',), readonly=True)
    max_volume = Parameter('max volume', FloatRange(0,100000, unit='uL',), readonly=True)
    
    target_flow_rate = Parameter('target flow rate', FloatRange(unit='uL/s'), readonly=False)
    real_flow_rate = Parameter('actual flow rate', FloatRange(unit='uL/s'), readonly=True)
    
    target = Parameter('target volume', FloatRange(unit='uL'), readonly=False)

    no_of_valve_pos = Property('number of valve positions', IntRange(0,10), default=1)
    valve_pos = Parameter('valve position', EnumType('valve', CLOSED=0, APP=1, RES=2, OPEN=3), readonly=False)

    force = Parameter('syringe force', FloatRange(unit='kN'), readonly=True)
    max_force = Parameter('max device force', FloatRange(unit='kN'), readonly=True)
    force_limit = Parameter('user force limit', FloatRange(unit='kN'), readonly=False)

    _resolving_force_overload = False

    def initModule(self):
        super().initModule()

        self.pump = qmixpump.Pump()
        self.pump.lookup_by_name(self.pump_name)

        self.valve = qmixvalve.Valve()
        self.valve.lookup_by_name(self.valve_name)


    def initialReads(self):
        if self.pump.is_in_fault_state():
            self.pump.clear_fault()
        
        if not self.pump.is_enabled():
            self.pump.enable(True)
        
        self.pump.set_syringe_param(self.inner_diameter_set, self.piston_stroke_set)

        self.pump.set_volume_unit(qmixpump.UnitPrefix.micro, qmixpump.VolumeUnit.litres)

        self.pump.set_flow_unit(qmixpump.UnitPrefix.micro, qmixpump.VolumeUnit.litres, qmixpump.TimeUnit.per_second)
        
        self.max_flow_rate = round(self.pump.get_flow_rate_max(),2)
        self.max_volume = round(self.pump.get_volume_max(),2)
        self.valve_pos = self.valve.actual_valve_position()

        self.target_flow_rate = round(self.max_flow_rate * 0.5,2)
        self.target = max(0, round(self.pump.get_fill_level(),2))

        self.pump.enable_force_monitoring(True)
        self.max_force = self.pump.get_max_device_force()
        self.force_limit = self.max_force
    
    def read_value(self):
        return round(self.pump.get_fill_level(),2)

    def write_target(self, target):
        if self.read_valve_pos() == 0 :
            self.status = ERROR, 'Cannot pump if valve is closed'
            self.log.warn('Cannot pump if valve is closed')
            return target
        else:
            self.pump.set_fill_level(target, self.target_flow_rate)
            self.status = BUSY, 'Target changed'
            self.log.info(f'Started pumping at {self.target_flow_rate} ul/s')
            return target

    def write_target_flow_rate(self, rate):
        self.target_flow_rate = rate
        return rate

    def read_real_flow_rate(self):
        return round(self.pump.get_flow_is(),2)

    def read_valve_pos(self):
        return self.valve.actual_valve_position()

    def write_valve_pos(self, target_pos):
        self.valve.switch_valve_to_position(target_pos)
        return target_pos

    def read_force(self):
        return round(self.pump.read_force_sensor(),3)

    def read_force_limit(self):
        return self.pump.get_force_limit()

    def write_force_limit(self, limit):
        self.pump.write_force_limit(limit)
        return limit

    def read_status(self):
        fault_state = self.pump.is_in_fault_state()
        pumping = self.pump.is_pumping()
        pump_enabled = self.pump.is_enabled()
        safety_stop_active = self.pump.is_force_safety_stop_active()
        
        if fault_state == True:
            return ERROR, 'Pump in fault state'
        elif self._resolving_force_overload :
            return BUSY, 'Resolving force overload'
        elif safety_stop_active:
            return ERROR, 'Pressure safety stop'
        elif not pump_enabled:
            return ERROR, 'Pump not enabled'
        elif pumping == True:
            return BUSY, f'Pumping {self.real_flow_rate} ul/s'
        elif self.read_valve_pos() == 0:
            return IDLE, 'Valve closed'
        else:
            return IDLE, ''

    @Command
    def stop(self):
        self.pump.stop_pumping()
        self.target = self.pump.get_fill_level()
        self.status = BUSY, 'Stopping'

    @Command
    def clear_errors(self):
        """Clear fault state and enable pump"""
        if self.pump.is_in_fault_state():
            self.pump.clear_fault()
            self.log.info('Cleared faults')
        
        if not self.pump.is_enabled():
            self.pump.enable(True)
            self.log.info('Pump was disabled, re-enabling')
        
        self.target = max(0,round(self.value,2))
        self.status = IDLE, ''

    @Command
    def resolve_force_overload(self):
        """Resolve a force overload situation"""
        if not self.pump.is_force_safety_stop_active():
            self.status = ERROR, 'No force overload detected'
            self.log.warn('No force overload to be resolved')
            return
        
        self._resolving_force_overload = True
        self.status = BUSY, 'Resolving force overload'
        self.pump.enable_force_monitoring(False)

        flow = 0  - self.pump.get_flow_rate_max() / 100
        self.pump.generate_flow(flow)

        safety_stop_active = False
        while not safety_stop_active:
            time.sleep(0.1)
            safety_stop_active = self.pump.is_force_safety_stop_active()
        self.pump.stop_pumping()
        self.pump.enable_force_monitoring(True)

        time.sleep(0.3)
        self._resolving_force_overload = False
        self.status = self.read_status()

class ContiFlowPump(Drivable):
    io = Attached()

    inner_diameter_set = Property('inner diameter', FloatRange(), default=1)
    piston_stroke_set = Property('piston stroke', FloatRange(), default=60)

    crossflow_seconds = Property('crossflow duration', FloatRange(unit='s'),default=2)
    
    value = PersistentParam('flow rate', FloatRange(unit='uL/s'))
    status = PersistentParam()

    max_refill_flow = Parameter('max refill flow', FloatRange(unit='uL/s'), readonly=True)
    refill_flow = Parameter('refill flow', FloatRange(unit='uL/s'), readonly=False)

    max_flow_rate = Parameter('max flow rate', FloatRange(0,100000, unit='uL/s',), readonly=True)
    target = Parameter('target flow rate', FloatRange(unit='uL/s'), readonly=False)

    def initModule(self):
        super().initModule()

        self.pump = qmixpump.ContiFlowPump()
        self.pump.lookup_by_name("ContiFlowPump_1")

    def initialReads(self):
        if self.pump.is_in_fault_state():
            self.pump.clear_fault()
        
        if not self.pump.is_enabled():
            self.pump.enable(True)

        self.syringe_pump1 = self.pump.get_syringe_pump(0)
        self.syringe_pump1.set_syringe_param(self.inner_diameter_set, self.piston_stroke_set)
        self.syringe_pump2 = self.pump.get_syringe_pump(1)
        self.syringe_pump2.set_syringe_param(self.inner_diameter_set, self.piston_stroke_set)
        self.pump.set_volume_unit(qmixpump.UnitPrefix.micro, qmixpump.VolumeUnit.litres)
        self.pump.set_flow_unit(qmixpump.UnitPrefix.micro, qmixpump.VolumeUnit.litres, qmixpump.TimeUnit.per_second)

        self.pump.set_device_property(ContiFlowProperty.SWITCHING_MODE, ContiFlowSwitchingMode.CROSS_FLOW)
        self.max_refill_flow = self.pump.get_device_property(ContiFlowProperty.MAX_REFILL_FLOW)
        self.pump.set_device_property(ContiFlowProperty.REFILL_FLOW, self.max_refill_flow / 2.0)
        self.pump.set_device_property(ContiFlowProperty.CROSSFLOW_DURATION_S, self.crossflow_seconds)
        self.pump.set_device_property(ContiFlowProperty.OVERLAP_DURATION_S, 0)
        
        self.max_flow_rate = self.pump.get_flow_rate_max()
        self.target = 0

    def read_value(self):
        return round(self.pump.get_flow_is(),3)

    def write_target(self, target):
        if target <= 0:
            self.pump.stop_pumping()
            self.status = self.read_status()
            return 0
        else:
            self.pump.generate_flow(target)
            self.status = BUSY, 'Target changed'
            return target

    def read_refill_flow(self):
        return round(self.pump.get_device_property(ContiFlowProperty.REFILL_FLOW),3)

    def write_refill_flow(self, refill_flow):
        self.pump.set_device_property(ContiFlowProperty.REFILL_FLOW, refill_flow)
        self.max_flow_rate = self.pump.get_flow_rate_max()
        return refill_flow

    def read_status(self):
        fault_state = self.pump.is_in_fault_state()
        pumping = self.pump.is_pumping()
        pump_enabled = self.pump.is_enabled()
        pump_initialised = self.pump.is_initialized()
        pump_initialising = self.pump.is_initializing()
        
        if fault_state == True:
            return ERROR, 'Pump in fault state'
        elif not pump_enabled:
            return ERROR, 'Pump not enabled'
        elif not pump_initialised:
            return WARN, 'Pump not initialised'
        elif pump_initialising:
            return BUSY, 'Pump initialising'
        elif pumping == True:
            return BUSY, 'Pumping'
        else:
            return IDLE, ''

    @Command
    def stop(self):
        self.pump.stop_pumping()
        self.target = 0
        self.status = BUSY, 'Stopping'

    @Command
    def clear_errors(self):
        """Clear fault state and enable pump"""
        if self.pump.is_in_fault_state():
            self.pump.clear_fault()
            self.log.info('Cleared faults')
        
        if not self.pump.is_enabled():
            self.pump.enable(True)
            self.log.info('Pump was disabled, re-enabling')
        
        self.target = 0
        self.status = IDLE, ''

    @Command
    def initialise(self):
        """Initialise the ConfiFlow pump"""
        self.pump.initialize()