# Author: Wouter Gruenewald<wouter.gruenewald@psi.ch>

from frappy.core import StringType, BoolType, EnumType, FloatRange, Parameter, Property, PersistentParam, Command, IDLE, ERROR, WARN, BUSY, Drivable


class PeristalticPump(Drivable):
    value = Parameter('Pump speed', FloatRange(0,100,unit="%"), default=0)
    target = Parameter('Target pump speed', FloatRange(0,100,unit="%"), default=0)
    status = Parameter()

    addr_AO = Property('Address of the analog out', StringType())
    addr_dir_relay = Property('Address of the direction relay', StringType())
    addr_run_relay = Property('Address of the running relay', StringType())

    direction = Parameter('pump direction', EnumType('direction', CLOCKWISE=0, ANTICLOCKWISE=1), default=0, readonly=False)
    active = Parameter('pump running', BoolType(), default=False, readonly=False)

    def initModule(self):
        super().initModule()
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO+'_enabled', 'w') as f :
            f.write('0')
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO+'_mode', 'w') as f :
            f.write('V')
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO, 'w') as f :
            f.write('0')
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO+'_enabled', 'w') as f :
            f.write('1')

    def shutdownModule(self):
        '''Disable analog output'''
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO, 'w') as f :
            f.write('0')
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO+'_enabled', 'w') as f :
            f.write('0')

    def read_value(self):
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO, 'r') as f :
            raw_value = f.read().strip('\n')
        value = (int(raw_value) / 5000) * 100
        return value

    def write_target(self, target):
        raw_value = (target / 100)*5000
        with open('/sys/class/ionopimax/analog_out/'+self.addr_AO, 'w') as f :
            f.write(str(int(raw_value)))
        return target

    def read_direction(self):
        with open('/sys/class/ionopimax/digital_out/'+self.addr_dir_relay, 'r') as f :
            raw_direction = f.read().strip('\n')
        if raw_direction == '0' or raw_direction == 'F':
            return 0
        if raw_direction == '1' or raw_direction == 'S':
            return 1
        else:
            return None

    def write_direction(self, direction):
        if direction == 0:
            raw_direction = '0'
        elif direction == 1:
            raw_direction = '1'
        with open('/sys/class/ionopimax/digital_out/'+self.addr_dir_relay, 'w') as f :
            f.write(raw_direction)
        return direction

    def read_active(self):
        with open('/sys/class/ionopimax/digital_out/'+self.addr_run_relay, 'r') as f :
            raw_active = f.read().strip('\n')
        if raw_active == '0' or raw_active == 'F':
            return False
        elif raw_active == '1' or raw_active == 'S':
            return True
        else:
            return None

    def write_active(self, active):
        if active == False:
            raw_active = '0'
        elif active == True:
            raw_active = '1'
        with open('/sys/class/ionopimax/digital_out/'+self.addr_run_relay, 'w') as f :
            f.write(raw_active)
        return active


    def read_status(self):
        with open('/sys/class/ionopimax/digital_out/'+self.addr_dir_relay, 'r') as f :
            raw_direction = f.read().strip('\n')
        with open('/sys/class/ionopimax/digital_out/'+self.addr_run_relay, 'r') as f :
            raw_active = f.read().strip('\n')

        if raw_direction == 'F' or raw_direction == 'S':
            return ERROR, 'Fault on direction relay'
        elif raw_active == 'F' or raw_active == 'S':
            return ERROR, 'Fault on pump activation relay'
        elif self.active == True:
            return BUSY, 'Pump running'
        else:
            return IDLE, ''

    @Command
    def stop(self):
        self.write_active(False)
